import polars_ols as pls
from polars import Expr, Int32, UInt16, map_batches
from polars import arange, repeat
from polars import rolling_corr, rolling_cov
from polars_ols import RollingKwargs

from polars_ta import TA_EPSILON
from polars_ta.utils.numba_ import batches_i1_o1, batches_i2_o1, roll_split_i2_o1
from polars_ta.utils.pandas_ import roll_kurt, roll_rank
from polars_ta.wq._nb import roll_argmax, roll_argmin, roll_prod, roll_co_kurtosis, roll_co_skewness, roll_moment, roll_partial_corr, roll_triple_corr, _zip_prod, _zip_sum


def ts_arg_max(x: Expr, d: int = 5, reverse: bool = True) -> Expr:
    return x.map_batches(lambda x1: batches_i1_o1(x1.to_numpy(), roll_argmax, d, reverse, dtype=UInt16))


def ts_arg_min(x: Expr, d: int = 5, reverse: bool = True) -> Expr:
    return x.map_batches(lambda x1: batches_i1_o1(x1.to_numpy(), roll_argmin, d, reverse, dtype=UInt16))


def ts_co_kurtosis(x: Expr, y: Expr, d: int = 5, ddof: int = 0) -> Expr:
    return map_batches([x, y], lambda xx: batches_i2_o1([x1.to_numpy() for x1 in xx], roll_co_kurtosis, d))


def ts_co_skewness(x: Expr, y: Expr, d: int = 5, ddof: int = 0) -> Expr:
    return map_batches([x, y], lambda xx: batches_i2_o1([x1.to_numpy() for x1 in xx], roll_co_skewness, d))


def ts_corr(x: Expr, y: Expr, d: int = 5, ddof: int = 1) -> Expr:
    # x, y is indifferent with y, x
    # x、y不区分先后
    return rolling_corr(x, y, window_size=d, ddof=ddof)


def ts_count(x: Expr, d: int = 30) -> Expr:
    return x.cast(Int32).rolling_sum(d)


def ts_count_nans(x: Expr, d: int = 5) -> Expr:
    # null or nan?
    # null与nan到底用哪一个？
    return x.is_null().rolling_sum(d)


def ts_covariance(x: Expr, y: Expr, d: int = 5, ddof: int = 1) -> Expr:
    # x, y is indifferent with y, x
    # x、y不区分先后
    return rolling_cov(x, y, window_size=d, ddof=ddof)


def ts_decay_exp_window(x: Expr, d: int = 30, factor: float = 1.0) -> Expr:
    # TODO weights not yet supported on array with null values
    y = arange(d - 1, -1, step=-1, eager=False)
    weights = repeat(factor, d, eager=True).pow(y)
    return x.rolling_mean(d, weights=weights)


def ts_decay_linear(x: Expr, d: int = 30) -> Expr:
    # TODO weights not yet supported on array with null values
    weights = arange(1, d + 1, eager=True)
    return x.rolling_mean(d, weights=weights)


def ts_delay(x: Expr, d: int = 1) -> Expr:
    return x.shift(d)


def ts_delta(x: Expr, d: int = 1) -> Expr:
    return x.diff(d)


def ts_ir(x: Expr, d: int = 1) -> Expr:
    return ts_mean(x, d) / ts_std_dev(x, d, 0)


def ts_kurtosis(x: Expr, d: int = 5) -> Expr:
    # TODO 等待polars官方出rolling_kurt
    return x.map_batches(lambda a: roll_kurt(a, d))


def ts_log_diff(x: Expr, d: int = 1) -> Expr:
    """Returns log(current value of input or x[t] ) - log(previous value of input or x[t-1])."""
    return x.log().diff(d)


def ts_max(x: Expr, d: int = 30) -> Expr:
    return x.rolling_max(d)


def ts_max_diff(x: Expr, d: int = 30) -> Expr:
    return x - ts_max(x, d)


def ts_mean(x: Expr, d: int = 5) -> Expr:
    return x.rolling_mean(d)


def ts_median(x: Expr, d: int = 5) -> Expr:
    return x.rolling_median(d)


def ts_min(x: Expr, d: int = 30) -> Expr:
    return x.rolling_min(d)


def ts_min_diff(x: Expr, d: int = 30) -> Expr:
    return x - ts_min(x, d)


def ts_min_max_cps(x: Expr, d: int, f: float = 2) -> Expr:
    """Returns (ts_min(x, d) + ts_max(x, d)) - f * x. If not specified, by default f = 2"""
    return (ts_min(x, d) + ts_max(x, d)) - f * x


def ts_min_max_diff(x: Expr, d: int, f: float = 0.5) -> Expr:
    """Returns x - f * (ts_min(x, d) + ts_max(x, d)). If not specified, by default f = 0.5"""
    return x - f * (ts_min(x, d) + ts_max(x, d))


def ts_moment(x: Expr, d: int, k: int = 0) -> Expr:
    """Returns K-th central moment of x for the past d days."""
    return x.map_batches(lambda x1: batches_i1_o1(x1.to_numpy(), roll_moment, d, k))


def ts_partial_corr(x: Expr, y: Expr, z: Expr, d: int) -> Expr:
    """Returns partial correlation of x, y, z for the past d days."""
    return map_batches([x, y, z], lambda xx: batches_i2_o1([x1.to_numpy() for x1 in xx], roll_partial_corr, d))


def ts_percentage(x: Expr, d: int, percentage: float = 0.5) -> Expr:
    """Returns percentile value of x for the past d days."""
    return x.rolling_quantile(percentage, window_size=d)


def ts_product(x: Expr, d: int = 5) -> Expr:
    return x.map_batches(lambda x1: batches_i1_o1(x1.to_numpy(), roll_prod, d))


def ts_rank(x: Expr, d: int = 5) -> Expr:
    # TODO 等待polars官方出rolling_rank，并支持pct
    # bottleneck长期无人维护，pydata/bottleneck#434 没有合并
    # pandas中已经用跳表实现了此功能，速度也不差
    return x.map_batches(lambda a: roll_rank(a, d, True))


def ts_returns(x: Expr, d: int = 1) -> Expr:
    return x.pct_change(d)


def ts_scale(x: Expr, d: int = 5) -> Expr:
    a = ts_min(x, d)
    b = ts_max(x, d)
    return (x - a) / (b - a + TA_EPSILON)


def ts_skewness(x: Expr, d: int = 5, bias: bool = False) -> Expr:
    # same with pandas when bias=False
    # bias=False与pandas结果一样
    return x.rolling_skew(d, bias=bias)


def ts_std_dev(x: Expr, d: int = 5, ddof: int = 0) -> Expr:
    return x.rolling_std(d, ddof=ddof)


def ts_sum(x: Expr, d: int = 30) -> Expr:
    return x.rolling_sum(d)


def ts_triple_corr(x: Expr, y: Expr, z: Expr, d: int) -> Expr:
    """Returns triple correlation of x, y, z for the past d days."""
    return map_batches([x, y, z], lambda xx: batches_i2_o1([x1.to_numpy() for x1 in xx], roll_triple_corr, d))


def ts_weighted_delay(x: Expr, k: float = 0.5) -> Expr:
    return x.rolling_sum(2, weights=[1 - k, k])


def ts_zscore(x: Expr, d: int = 5) -> Expr:
    return (x - ts_mean(x, d)) / ts_std_dev(x, d)


def ts_zip_prod(v: Expr, r: Expr) -> Expr:
    """z字形累乘。可用于市值累乘日收益率的情况，如成份股权重按行情更新"""
    return map_batches([v, r], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _zip_prod))


def ts_zip_sum(x: Expr, y: Expr) -> Expr:
    """z字形累加"""
    return map_batches([x, y], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _zip_sum))


def ts_regression_resid(y: Expr, x: Expr, d: int) -> Expr:
    return pls.compute_rolling_least_squares(y, x, mode='residuals', add_intercept=True, rolling_kwargs=RollingKwargs(window_size=d, min_periods=d))


def ts_regression_pred(y: Expr, x: Expr, d: int) -> Expr:
    return pls.compute_rolling_least_squares(y, x, mode='predictions', add_intercept=True, rolling_kwargs=RollingKwargs(window_size=d, min_periods=d))


def ts_regression_intercept(y: Expr, x: Expr, d: int) -> Expr:
    return pls.compute_rolling_least_squares(y, x, mode='coefficients', add_intercept=True, rolling_kwargs=RollingKwargs(window_size=d, min_periods=d)).struct[1]


def ts_regression_slope(y: Expr, x: Expr, d: int) -> Expr:
    return pls.compute_rolling_least_squares(y, x, mode='coefficients', add_intercept=True, rolling_kwargs=RollingKwargs(window_size=d, min_periods=d)).struct[0]


def ts_resid(y: Expr, *more_x: Expr, d: int) -> Expr:
    """多元时序滚动回归取残差"""
    return pls.compute_rolling_least_squares(y, *more_x, mode='residuals', rolling_kwargs=RollingKwargs(window_size=d, min_periods=d))


def ts_pred(y: Expr, *more_x: Expr, d: int) -> Expr:
    """多元时序滚动回归预测"""
    return pls.compute_rolling_least_squares(y, *more_x, mode='predictions', rolling_kwargs=RollingKwargs(window_size=d, min_periods=d))


def ts_weighted_mean(x: Expr, w: Expr, d: int) -> Expr:
    """时序加权平均"""
    return (x * w).rolling_sum(d) / w.rolling_sum(d)


def ts_weighted_sum(x: Expr, w: Expr, d: int) -> Expr:
    """时序加权求和"""
    return (x * w).rolling_sum(d)


def ts_split_sum(a: Expr, b: Expr, d: int, n: int) -> Expr:
    """切割论求和，以b为依据，对a进行切割

    在d窗口范围内以b为依据进行从小到大排序。最大的N个和-最小的N个和
    """
    return roll_split_i2_o1(a, b, d, n)



from polars import Expr, max_horizontal

from polars_ta.ta.overlap import RMA


def ATR(high: Expr, low: Expr, close: Expr, timeperiod: int = 14) -> Expr:
    """"""
    return RMA(TRANGE(high, low, close), timeperiod)


def NATR(high: Expr, low: Expr, close: Expr, timeperiod: int = 14) -> Expr:
    """

    Notes
    -----
    talib.ATR multiples another 100
    talib.ATR版相当于多乘了100

    """
    return ATR(high, low, close, timeperiod) / close


def TRANGE(high: Expr, low: Expr, close: Expr) -> Expr:
    """

    Notes
    -----
    the 0-th position is `x` rather than `nan` in talib
    第0位置为max(x,nan,nan)=x,比talib多一个值

    """
    prev_close = close.shift(1)
    tr1 = high - low
    tr2 = (high - prev_close).abs()
    tr3 = (low - prev_close).abs()
    return max_horizontal(tr1, tr2, tr3)



"""
Demo for using numba to implement rolling functions.
本文件是使用numba实现rolling的函数，演示用
"""
from typing import List

import numpy as np
from numba import jit
from numpy.lib.stride_tricks import sliding_window_view
from polars import Series, Expr, map_batches

"""
Series.to_numpy的操作在调用之前做，这样可控一些
batches_i1_o1这一类的函数输入不支持Series，只支持numpy。设计成在map_batches转换更可控
"""


def batches_i1_o1(x1: np.ndarray, func, *args, dtype=None) -> Series:
    return Series(func(x1, *args), nan_to_null=True, dtype=dtype)


def batches_i2_o1(xx: List[np.ndarray], func, *args, dtype=None) -> Series:
    return Series(func(*xx, *args), nan_to_null=True, dtype=dtype)


def batches_i1_o2(x1: np.ndarray, func, *args, dtype=None, ret_idx: int = 0) -> Series:
    return Series(func(x1, *args)[ret_idx], nan_to_null=True, dtype=dtype)


def batches_i2_o2(xx: List[np.ndarray], func, *args, dtype=None, ret_idx: int = 0) -> Series:
    return Series(func(*xx, *args)[ret_idx], nan_to_null=True, dtype=dtype)


@jit(nopython=True, nogil=True, cache=True)
def nb_roll_sum(x1, window):
    """Demo code. Use `pl.col('A').rolling_sum(10).alias('a1')` instead.
    演示代码，请直接用 pl.col('A').rolling_sum(10).alias('a1')"""
    out = np.full(x1.shape, np.nan, dtype=float)
    if len(x1) < window:
        return out
    a1 = sliding_window_view(x1, window)
    for i, v1 in enumerate(a1):
        out[i + window - 1] = np.sum(v1)
    return out


@jit(nopython=True, nogil=True, cache=True)
def nb_roll_cov(x1, x2, window):
    """Demo code. Use `pl.rolling_cov(pl.col('A'), pl.col('B'), window_size=10).alias('a6')` instead.
    演示代码，pl.rolling_cov(pl.col('A'), pl.col('B'), window_size=10).alias('a6')"""
    out = np.full(x1.shape, np.nan, dtype=float)
    if len(x1) < window:
        return out
    a1 = sliding_window_view(x1, window)
    a2 = sliding_window_view(x2, window)
    for i, (v1, v2) in enumerate(zip(a1, a2)):
        out[i + window - 1] = np.cov(v1, v2)[0, 1]
    return out


def roll_sum(x: Expr, n: int) -> Expr:
    return x.map_batches(lambda x1: batches_i1_o1(x1.to_numpy(), nb_roll_sum, n))


def roll_cov(a: Expr, b: Expr, n: int) -> Expr:
    return map_batches([a, b], lambda xx: batches_i2_o1([x1.to_numpy() for x1 in xx], nb_roll_cov, n))


@jit(nopython=True, nogil=True, cache=True)
def nb_split_o1(x1, x2, window=10, n=2):
    out = np.full((x1.shape[0], 2), np.nan, dtype=float)
    if len(x1) < window:
        return out[:, 1] - out[:, 0]
    a1 = sliding_window_view(x1, window)
    a2 = sliding_window_view(x2, window)
    for i, (v1, v2) in enumerate(zip(a1, a2)):
        # 排序两次，解决nan的问题
        b1 = np.argsort(v2)[:n]
        b2 = np.argsort(-v2)[:n]
        out[i + window - 1, 0] = np.sum(v1[b1])
        out[i + window - 1, 1] = np.sum(v1[b2])
    return out[:, 1] - out[:, 0]


@jit(nopython=True, nogil=True, cache=True)
def nb_split_o2(x1, x2, window=10, n=2):
    out = np.full((x1.shape[0], 2), np.nan, dtype=float)
    if len(x1) < window:
        return out
    a1 = sliding_window_view(x1, window)
    a2 = sliding_window_view(x2, window)
    for i, (v1, v2) in enumerate(zip(a1, a2)):
        # 排序两次，解决nan的问题
        b1 = np.argsort(v2)[:n]
        b2 = np.argsort(-v2)[:n]
        out[i + window - 1, 0] = np.sum(v1[b1])
        out[i + window - 1, 1] = np.sum(v1[b2])
    return out


def roll_split_i2_o1(a: Expr, b: Expr, d: int, n: int) -> Expr:
    """切割论的示例。只输出一个结果"""
    return map_batches(
        [a, b],
        lambda xx: batches_i2_o1([x1.to_numpy() for x1 in xx], nb_split_o1, d, n)
    )


def roll_split_i2_o2(a: Expr, b: Expr, d: int, n: int) -> Expr:
    """切割论的示例。两输出，由用户自己进行后期处理"""
    # !!!输出为struct时的特殊写法，使用其它写法都会导致之后想struct.field失败
    return map_batches(
        [a, b],
        lambda xx: batches_i2_o1([x1.to_numpy() for x1 in xx], nb_split_o2, d, n).list.to_struct()
    ).struct.rename_fields(['split_a', 'split_b'])



"""
In this file we use
open_: Expr, high: Expr, low: Expr, close: Expr
for all parameters
to ensure the similar function signatures when calling them

本文件中参数全用
open_: Expr, high: Expr, low: Expr, close: Expr
统一的好处是在使用时不用考虑函数调用区别
"""
from polars import Expr

from polars_ta.candles.cdl1 import lower_body, upper_body


# https://github.com/TA-Lib/ta-lib/blob/main/src/ta_func/ta_utility.h#L360
def ts_gap_up(open_: Expr, high: Expr, low: Expr, close: Expr) -> Expr:
    """跳空高开"""
    return low > high.shift(1)


def ts_gap_down(open_: Expr, high: Expr, low: Expr, close: Expr) -> Expr:
    """跳空低开"""
    return high < low.shift(1)


def ts_real_body_gap_up(open_: Expr, high: Expr, low: Expr, close: Expr) -> Expr:
    """实体跳空高开"""
    return lower_body(open_, high, low, close) > upper_body(open_, high, low, close).shift(1)


def ts_real_body_gap_down(open_: Expr, high: Expr, low: Expr, close: Expr) -> Expr:
    """实体跳空低开"""
    return upper_body(open_, high, low, close) < lower_body(open_, high, low, close).shift(1)



from functools import lru_cache
from typing import Tuple

import numpy as np
from numba import jit
from pandas._libs.window.aggregations import roll_kurt as _roll_kurt
from pandas._libs.window.aggregations import roll_rank as _roll_rank
from polars import Series

"""
When converting float32 to float64 before computing. Either use
x.cast(Float64).to_numpy()
or
x.to_numpy().astype(float)

The second one is faster

在计算前需要将float32转成float64，有以下两种方法
x.cast(Float64).to_numpy()
x.to_numpy().astype(float)

第二种方法更快
"""


@lru_cache
@jit(nopython=True, nogil=True, fastmath=True, cache=True)
def get_window_bounds(
        num_values: int = 0,
        window_size: int = 10,
) -> Tuple[np.ndarray, np.ndarray]:
    end = np.arange(1, num_values + 1, dtype=np.int64)
    start = end - window_size
    start = np.clip(start, 0, num_values)
    return start, end


def roll_rank(x: Series, d: int, pct: bool = True, method: str = 'average', ascending: bool = True):
    start, end = get_window_bounds(len(x), d)
    """
    https://github.com/pandas-dev/pandas/blob/main/pandas/_libs/window/aggregations.pyx#L1281

    def roll_rank(const float64_t[:] values, ndarray[int64_t] start,
              ndarray[int64_t] end, int64_t minp, bint percentile,
              str method, bint ascending) -> np.ndarray:

    O(N log(window)) implementation using skip list
    """
    ret = _roll_rank(x.to_numpy().astype(float), start, end, d, pct, method, ascending)
    return Series(ret, nan_to_null=True)


def roll_kurt(x, d):
    start, end = get_window_bounds(len(x), d)
    """
    https://github.com/pandas-dev/pandas/blob/main/pandas/_libs/window/aggregations.pyx#L803

    def roll_kurt(ndarray[float64_t] values, ndarray[int64_t] start,
              ndarray[int64_t] end, int64_t minp) -> np.ndarray:
    """
    ret = _roll_kurt(x.to_numpy().astype(float), start, end, d)
    return Series(ret, nan_to_null=True)



from typing import List

import numpy as np
import polars_ols as pls
from polars import Expr, Series, Struct, map_batches
from polars_ols.least_squares import OLSKwargs

from polars_ta import TA_EPSILON
from polars_ta.wq.cross_sectional import cs_rank


# ======================
# standardize
def cs_zscore(x: Expr, ddof: int = 0) -> Expr:
    return (x - x.mean()) / x.std(ddof=ddof)


def cs_minmax(x: Expr) -> Expr:
    a = x.min()
    b = x.max()
    return (x - a) / (b - a + TA_EPSILON)


# ======================
# winsorize
def cs_quantile(x: Expr, low_limit: float = 0.025, up_limit: float = 0.995) -> Expr:
    a = x.quantile(low_limit)
    b = x.quantile(up_limit)
    return x.clip(lower_bound=a, upper_bound=b)


def cs_3sigma(x: Expr, n: float = 3.) -> Expr:
    # fill_nan will seriously reduce speed. So it's more appropriate for users to handle it themselves
    # fill_nan(None) 严重拖慢速度，所以还是由用户自己处理更合适
    a = x.mean()
    b = n * x.std(ddof=0)
    return x.clip(lower_bound=a - b, upper_bound=a + b)


def cs_mad(x: Expr, n: float = 3., k: float = 1.4826) -> Expr:
    # https://en.wikipedia.org/wiki/Median_absolute_deviation
    a = x.median()
    b = (n * k) * (x - a).abs().median()
    return x.clip(lower_bound=a - b, upper_bound=a + b)


# ======================
# neutralize
def cs_demean(x: Expr) -> Expr:
    """demean

    Notes
    -----
    Slower than multivariate regression. We need to groupby date and industry here,
    while multivariate regression only needs to add industry dummy variables and then groupby date

    Notes
    -----
    速度没有多元回归快，因为这里需要按日期行业groupby，
    而多元回归只要添加行业哑变量，然后按日期groupby即可
    """
    return x - x.mean()


# ======================
# neutralize
_ols_kwargs = OLSKwargs(null_policy='drop', solve_method='svd')


def _residual_multiple(cols: List[Series], add_constant: bool) -> Series:
    # 将pl.Struct转成list,这样可以实现传正则，其它也转list
    cols = [list(c.struct) if isinstance(c.dtype, Struct) else [c] for c in cols]
    # 二维列表转一维列表，再转np.ndarray
    cols = [i.to_numpy() for p in cols for i in p]
    if add_constant:
        cols += [np.ones_like(cols[0])]
    yx = np.vstack(cols).T

    # skip nan
    mask = np.any(np.isnan(yx), axis=1)
    yx_ = yx[~mask, :]

    y = yx_[:, 0]
    x = yx_[:, 1:]
    coef = np.linalg.lstsq(x, y, rcond=None)[0]
    y_hat = np.sum(x * coef, axis=1)
    residual = y - y_hat

    # refill
    out = np.empty_like(yx[:, 0])
    out[~mask] = residual
    out[mask] = np.nan
    return Series(out, nan_to_null=True)


def cs_resid_(y: Expr, *more_x: Expr) -> Expr:
    """multivariate regression
    多元回归
    """
    return map_batches([y, *more_x], lambda xx: _residual_multiple(xx, False))


def cs_resid(y: Expr, *more_x: Expr) -> Expr:
    """多元回归取残差"""
    return pls.compute_least_squares(y, *more_x, mode='residuals', ols_kwargs=_ols_kwargs)


def cs_mad_zscore(y: Expr) -> Expr:
    """常用功能简化封装。去极值、标准化"""
    return cs_zscore(cs_mad(y))


def cs_mad_zscore_resid(y: Expr, *more_x: Expr) -> Expr:
    """常用功能简化封装。去极值、标准化、中性化"""
    return cs_resid(cs_zscore(cs_mad(y)), *more_x)


def cs_mad_rank(y: Expr) -> Expr:
    """常用功能简化封装。去极值，排名。

    适合于分层收益V型或倒V的情况"""
    return cs_rank(cs_mad(y))


def cs_mad_rank2(y: Expr, m: float) -> Expr:
    """非线性处理。去极值，排名，移动峰或谷到零点，然后平方

    适合于分层收益V型或倒V的情况"""
    return (cs_rank(cs_mad(y)) - m) ** 2


def cs_mad_rank2_resid(y: Expr, m: float, *more_x: Expr) -> Expr:
    """非线性处理。去极值，排名，移动峰或谷到零点，然后平方。回归取残差

    适合于分层收益V型或倒V的情况"""
    return cs_resid((cs_rank(cs_mad(y)) - m) ** 2, *more_x)


def cs_rank2(y: Expr, m: float) -> Expr:
    """非线性处理。移动峰或谷到零点，然后平方

    适合于分层收益V型或倒V的情况"""
    return (cs_rank(y) - m) ** 2



from polars import Expr

from polars_ta.tdx.choice import IF
from polars_ta.tdx.reference import REF
from polars_ta.tdx.reference import SUM
from polars_ta.tdx.reference import CUMSUM


def OBV(CLOSE: Expr, VOL: Expr) -> Expr:
    """
    VA:=IF(CLOSE>REF(CLOSE,1),VOL,-VOL);
    OBV:SUM(IF(CLOSE=REF(CLOSE,1),0,VA),0);
    MAOBV:MA(OBV,M);
    """
    LC = REF(CLOSE, 1)
    VA = IF(CLOSE - LC, VOL, -VOL)
    return CUMSUM(IF(CLOSE == LC, 0, VA))


def VR(CLOSE: Expr, VOL: Expr, N: int = 26) -> Expr:
    """
    TH:=SUM(IF(CLOSE>REF(CLOSE,1),VOL,0),N);
    TL:=SUM(IF(CLOSE<REF(CLOSE,1),VOL,0),N);
    TQ:=SUM(IF(CLOSE=REF(CLOSE,1),VOL,0),N);
    VR:100*(TH*2+TQ)/(TL*2+TQ);
    MAVR:MA(VR,M);
    """
    LC = REF(CLOSE, 1)
    TH = SUM(IF(CLOSE > LC, VOL, 0), N)
    TL = SUM(IF(CLOSE < LC, VOL, 0), N)
    TQ = SUM(IF(CLOSE == LC, VOL, 0), N)

    return (TH * 2 + TQ) / (TL * 2 + TQ)  # * 100



from polars import Series, Expr


def _avedev(x: Series) -> Series:
    # x is Series rather than Expr. Making this function slow
    # 可惜rolling_map后这里已经由Expr变成了Series
    return (x - x.mean()).abs().mean()


def AVEDEV(close: Expr, timeperiod: int = 5) -> Expr:
    """Avoid using this function, it is slow.
    平均绝对偏差
    """
    return close.rolling_map(_avedev, timeperiod)



from polars import Expr


# 本文件中算子都是对整列计算得到一个值，使用时可能会广播到整列
# 可以用在横截面，也可以用在时序。但时序使用会引入未来数据，这时一般用途是统计

def vec_avg(x: Expr) -> Expr:
    """Taking mean of the vector field x"""
    return x.mean()


def vec_choose(x: Expr, nth: int) -> Expr:
    """Choosing kth item(indexed at 0) from each vector field x"""
    return x.gather(nth)


def vec_count(x: Expr) -> Expr:
    """Number of elements in vector field x"""
    return x.count()


def vec_ir(x: Expr) -> Expr:
    """Information Ratio (Mean / Standard Deviation) of vector field x"""
    return x.mean() / x.std(ddof=0)


def vec_kurtosis(x: Expr) -> Expr:
    """Kurtosis of vector field x"""
    return x.kurtosis()


def vec_max(x: Expr) -> Expr:
    """Maximum value form vector field x"""
    return x.max()


def vec_min(x: Expr) -> Expr:
    """Minimum value form vector field x"""
    return x.min()


def vec_norm(x: Expr) -> Expr:
    """Sum of all absolute values of vector field x"""
    return x.abs().sum()


def vec_percentage(x: Expr, percentage: float = 0.5) -> Expr:
    """Percentile of vector field x"""
    return x.quantile(percentage)


def vec_powersum(x: Expr, constant: float = 2) -> Expr:
    """Sum of power of vector field x"""
    return (x ** constant).sum()


def vec_range(x: Expr) -> Expr:
    """Difference between maximum and minimum element in vector field x"""
    return x.max() - x.min()


def vec_skewness(x: Expr) -> Expr:
    """Skewness of vector field x"""
    return x.skew()


def vec_stddev(x: Expr) -> Expr:
    """Standard Deviation of vector field x"""
    return x.std(ddof=0)


def vec_sum(x: Expr) -> Expr:
    """Sum of vector field x"""
    return x.sum()



import numpy as np
from numba import jit
from numpy import argmax, argmin, prod, mean, std, full, vstack, corrcoef
from numpy.lib.stride_tricks import sliding_window_view


@jit(nopython=True, nogil=True, fastmath=True, cache=True)
def roll_argmax(x1, window, reverse):
    out = full(x1.shape, np.nan, dtype=float)
    if len(x1) < window:
        return out
    a1 = sliding_window_view(x1, window)
    if reverse:
        a1 = a1[:, ::-1]
    for i, v1 in enumerate(a1):
        out[i + window - 1] = argmax(v1)
    return out


@jit(nopython=True, nogil=True, fastmath=True, cache=True)
def roll_argmin(x1, window, reverse):
    out = full(x1.shape, np.nan, dtype=float)
    if len(x1) < window:
        return out
    a1 = sliding_window_view(x1, window)
    if reverse:
        a1 = a1[:, ::-1]
    for i, v1 in enumerate(a1):
        out[i + window - 1] = argmin(v1)
    return out


@jit(nopython=True, nogil=True, fastmath=True, cache=True)
def roll_prod(x1, window):
    out = full(x1.shape, np.nan, dtype=float)
    if len(x1) < window:
        return out
    a1 = sliding_window_view(x1, window)
    for i, v1 in enumerate(a1):
        out[i + window - 1] = prod(v1)
    return out


@jit(nopython=True, nogil=True, fastmath=True, cache=True)
def _co_kurtosis(a1, a2):
    t1 = a1 - mean(a1)
    t2 = a2 - mean(a2)
    t3 = std(a1)
    t4 = std(a2)
    return mean(t1 * (t2 ** 3)) / (t3 * (t4 ** 3))


@jit(nopython=True, nogil=True, fastmath=True, cache=True)
def roll_co_kurtosis(x1, x2, window):
    out = full(x1.shape, np.nan, dtype=float)
    if len(x1) < window:
        return out
    a1 = sliding_window_view(x1, window)
    a2 = sliding_window_view(x2, window)
    for i, (v1, v2) in enumerate(zip(a1, a2)):
        out[i + window - 1] = _co_kurtosis(v1, v2)
    return out


@jit(nopython=True, nogil=True, fastmath=True, cache=True)
def _co_skewness(a1, a2):
    t1 = a1 - mean(a1)
    t2 = a2 - mean(a2)
    t3 = std(a1)
    t4 = std(a2)
    return mean(t1 * (t2 ** 2)) / (t3 * (t4 ** 2))


@jit(nopython=True, nogil=True, fastmath=True, cache=True)
def roll_co_skewness(x1, x2, window):
    out = full(x1.shape, np.nan, dtype=float)
    if len(x1) < window:
        return out
    a1 = sliding_window_view(x1, window)
    a2 = sliding_window_view(x2, window)
    for i, (v1, v2) in enumerate(zip(a1, a2)):
        out[i + window - 1] = _co_skewness(v1, v2)
    return out


@jit(nopython=True, nogil=True, fastmath=True, cache=True)
def _moment(a1, k):
    """centeral moment
    中心矩阵"""
    return mean((a1 - mean(a1)) ** k)


@jit(nopython=True, nogil=True, fastmath=True, cache=True)
def roll_moment(x1, window, k):
    out = full(x1.shape, np.nan, dtype=float)
    if len(x1) < window:
        return out
    a1 = sliding_window_view(x1, window)
    for i, v1 in enumerate(a1):
        out[i + window - 1] = _moment(v1, k)
    return out


@jit(nopython=True, nogil=True, fastmath=True, cache=True)
def _partial_corr(a1, a2, a3):
    """TODO 不知是否正确，需要检查"""
    c = corrcoef(vstack((a1, a2, a3)))
    rxy = c[0, 1]
    rxz = c[0, 2]
    ryz = c[1, 2]
    t1 = rxy - rxz * ryz
    t2 = (1 - rxz ** 2) ** 0.5
    t3 = (1 - ryz ** 2) ** 0.5
    return t1 / (t2 * t3)


@jit(nopython=True, nogil=True, fastmath=True, cache=True)
def roll_partial_corr(x1, x2, x3, window):
    out = full(x1.shape, np.nan, dtype=float)
    if len(x1) < window:
        return out
    a1 = sliding_window_view(x1, window)
    a2 = sliding_window_view(x2, window)
    a3 = sliding_window_view(x3, window)
    for i, (v1, v2, v3) in enumerate(zip(a1, a2, a3)):
        out[i + window - 1] = _partial_corr(v1, v2, v3)
    return out


@jit(nopython=True, nogil=True, fastmath=True, cache=True)
def _triple_corr(a1, a2, a3):
    t1 = a1 - mean(a1)
    t2 = a2 - mean(a2)
    t3 = a3 - mean(a3)
    t4 = std(a1)
    t5 = std(a2)
    t6 = std(a3)
    return mean(t1 * t2 * t3) / (t4 * t5 * t6)


@jit(nopython=True, nogil=True, fastmath=True, cache=True)
def roll_triple_corr(x1, x2, x3, window):
    out = full(x1.shape, np.nan, dtype=float)
    if len(x1) < window:
        return out
    a1 = sliding_window_view(x1, window)
    a2 = sliding_window_view(x2, window)
    a3 = sliding_window_view(x3, window)
    for i, (v1, v2, v3) in enumerate(zip(a1, a2, a3)):
        out[i + window - 1] = _triple_corr(v1, v2, v3)
    return out


@jit(nopython=True, nogil=True, fastmath=True, cache=True)
def isnan(x):
    # https://github.com/numba/numba/issues/2919#issuecomment-747377615
    if int(x) == -9223372036854775808:
        return True
    else:
        return False


@jit(nopython=True, nogil=True, fastmath=True, cache=True)
def _zip_prod(a, b):
    for i in range(1, a.shape[0]):
        if isnan(a[i]):
            a[i] = a[i - 1] * b[i - 1]
    return a


@jit(nopython=True, nogil=True, fastmath=False, cache=False)
def _zip_sum(a, b):
    for i in range(1, a.shape[0]):
        if isnan(a[i]):
            a[i] = a[i - 1] + b[i - 1]
    return a



from polars import Expr

from polars_ta.wq.time_series import ts_corr
from polars_ta.wq.time_series import ts_std_dev


def BETA(high: Expr, low: Expr, timeperiod: int = 5) -> Expr:
    raise


def CORREL(high: Expr, low: Expr, timeperiod: int = 30) -> Expr:
    return ts_corr(high, low, timeperiod, 1)


def LINEARREG(close: Expr, timeperiod: int = 14) -> Expr:
    raise


def LINEARREG_ANGLE(close: Expr, timeperiod: int = 14) -> Expr:
    raise


def LINEARREG_INTERCEPT(close: Expr, timeperiod: int = 14) -> Expr:
    raise


def LINEARREG_SLOPE(close: Expr, timeperiod: int = 14) -> Expr:
    raise


def STDDEV(close: Expr, timeperiod: int = 5, nbdev: float = 1) -> Expr:
    return ts_std_dev(close, timeperiod, ddof=0) * nbdev


def TSF(close: Expr, timeperiod: int = 14) -> Expr:
    raise


def VAR(close: Expr, timeperiod: int = 5, nbdev: float = 1) -> Expr:
    return close.rolling_var(timeperiod, ddof=0) * nbdev



from polars import Expr

from polars_ta import TA_EPSILON
from polars_ta.tdx.arithmetic import ABS
from polars_ta.tdx.choice import IF
from polars_ta.tdx.reference import MA
from polars_ta.tdx.reference import REF
from polars_ta.tdx.reference import SUM
from polars_ta.tdx.reference import TR


def DPO(CLOSE: Expr, N: int = 20) -> Expr:
    """
    DPO:CLOSE-REF(MA(CLOSE,N),N/2+1);
    MADPO:MA(DPO,M);
    """

    return CLOSE - REF(MA(CLOSE, N), N // 2 + 1)


def EMV(HIGH: Expr, LOW: Expr, VOL: Expr, N: int = 14) -> Expr:
    """
    VOLUME:=MA(VOL,N)/VOL;
    MID:=100*(HIGH+LOW-REF(HIGH+LOW,1))/(HIGH+LOW);
    EMV:MA(MID*VOLUME*(HIGH-LOW)/MA(HIGH-LOW,N),N);
    MAEMV:MA(EMV,M);
    """
    ADD = HIGH + LOW
    SUB = HIGH - LOW

    VOLUME = MA(VOL, N) / VOL
    MID = 100 * (ADD - REF(ADD, 1)) / ADD
    return MA(MID * VOLUME * SUB / (MA(SUB, N) + TA_EPSILON), N)


def PLUS_DM(HIGH: Expr, LOW: Expr, N: int = 14) -> Expr:
    """
    MTR:=SUM(MAX(MAX(HIGH-LOW,ABS(HIGH-REF(CLOSE,1))),ABS(REF(CLOSE,1)-LOW)),N);
    HD :=HIGH-REF(HIGH,1);
    LD :=REF(LOW,1)-LOW;
    DMP:=SUM(IF(HD>0&&HD>LD,HD,0),N);
    DMM:=SUM(IF(LD>0&&LD>HD,LD,0),N);
    """
    HD = HIGH - REF(HIGH, 1)
    LD = REF(LOW, 1) - LOW
    DMP = SUM(IF((HD > 0) & (HD > LD), HD, 0), N)
    return DMP


def MINUS_DM(HIGH: Expr, LOW: Expr, N: int = 14) -> Expr:
    HD = HIGH - REF(HIGH, 1)
    LD = REF(LOW, 1) - LOW
    DMM = SUM(IF((LD > 0) & (LD > HD), LD, 0), N)
    return DMM


def PLUS_DI(HIGH: Expr, LOW: Expr, CLOSE: Expr, N: int = 14) -> Expr:
    """
    MTR:=SUM(MAX(MAX(HIGH-LOW,ABS(HIGH-REF(CLOSE,1))),ABS(REF(CLOSE,1)-LOW)),N);
    HD :=HIGH-REF(HIGH,1);
    LD :=REF(LOW,1)-LOW;
    DMP:=SUM(IF(HD>0&&HD>LD,HD,0),N);
    DMM:=SUM(IF(LD>0&&LD>HD,LD,0),N);
    PDI: DMP*100/MTR;
    MDI: DMM*100/MTR;
    """
    MTR = SUM(TR(HIGH, LOW, CLOSE), N)
    DMP = PLUS_DM(HIGH, LOW, N)
    PDI = DMP / MTR
    return PDI  # * 100


def MINUS_DI(HIGH: Expr, LOW: Expr, CLOSE: Expr, N: int = 14) -> Expr:
    MTR = SUM(TR(HIGH, LOW, CLOSE), N)
    DMM = MINUS_DM(HIGH, LOW, N)
    MDI = DMM / MTR
    return MDI  # * 100


def ADX(HIGH: Expr, LOW: Expr, CLOSE: Expr, N: int = 14, M: int = 6) -> Expr:
    """
    MTR:=SUM(MAX(MAX(HIGH-LOW,ABS(HIGH-REF(CLOSE,1))),ABS(REF(CLOSE,1)-LOW)),N);
    HD :=HIGH-REF(HIGH,1);
    LD :=REF(LOW,1)-LOW;
    DMP:=SUM(IF(HD>0&&HD>LD,HD,0),N);
    DMM:=SUM(IF(LD>0&&LD>HD,LD,0),N);
    PDI: DMP*100/MTR;
    MDI: DMM*100/MTR;
    ADX: MA(ABS(MDI-PDI)/(MDI+PDI)*100,M);
    ADXR:(ADX+REF(ADX,M))/2;
    """
    MTR = SUM(TR(HIGH, LOW, CLOSE), N)
    HD = HIGH - REF(HIGH, 1)
    LD = REF(LOW, 1) - LOW
    DMP = SUM(IF((HD > 0) & (HD > LD), HD, 0), N)
    DMM = SUM(IF((LD > 0) & (LD > HD), LD, 0), N)
    PDI = DMP / MTR  # * 100
    MDI = DMM / MTR  # * 100
    return MA(ABS(MDI - PDI) / (MDI + PDI), M)  # * 100


def ADXR(HIGH: Expr, LOW: Expr, CLOSE: Expr, N: int = 14, M: int = 6) -> Expr:
    adx = ADX(HIGH, LOW, CLOSE, N, M)
    adxr = (adx + REF(adx, M)) / 2
    return adxr



from polars_ta.wq.arithmetic import arc_cos as ACOS  # noqa
from polars_ta.wq.arithmetic import arc_sin as ASIN  # noqa
from polars_ta.wq.arithmetic import arc_tan as ATAN  # noqa
from polars_ta.wq.arithmetic import ceiling as CEIL  # noqa
from polars_ta.wq.arithmetic import cos as COS  # noqa
from polars_ta.wq.arithmetic import cosh as COSH  # noqa
from polars_ta.wq.arithmetic import exp as EXP  # noqa
from polars_ta.wq.arithmetic import floor as FLOOR  # noqa
from polars_ta.wq.arithmetic import log as LN  # noqa
from polars_ta.wq.arithmetic import log10 as LOG10  # noqa
from polars_ta.wq.arithmetic import sin as SIN  # noqa
from polars_ta.wq.arithmetic import sinh as SINH  # noqa
from polars_ta.wq.arithmetic import sqrt as SQRT  # noqa
from polars_ta.wq.arithmetic import tan as TAN  # noqa
from polars_ta.wq.arithmetic import tanh as TANH  # noqa



# generated by codegen_talib2.py
import talib as _ta
from polars import Expr, map_batches

from polars_ta.utils.numba_ import batches_i1_o1, batches_i1_o2, batches_i2_o1, batches_i2_o2


def HT_DCPERIOD(close: Expr) -> Expr:  # ['real']
    return close.map_batches(lambda x1: batches_i1_o1(x1.to_numpy().astype(float), _ta.HT_DCPERIOD))


def HT_DCPHASE(close: Expr) -> Expr:  # ['real']
    return close.map_batches(lambda x1: batches_i1_o1(x1.to_numpy().astype(float), _ta.HT_DCPHASE))


def HT_PHASOR(close: Expr, ret_idx: int = 1) -> Expr:  # ['inphase', 'quadrature']
    return close.map_batches(lambda x1: batches_i1_o2(x1.to_numpy().astype(float), _ta.HT_PHASOR, ret_idx=ret_idx))


def HT_SINE(close: Expr, ret_idx: int = 1) -> Expr:  # ['sine', 'leadsine']
    return close.map_batches(lambda x1: batches_i1_o2(x1.to_numpy().astype(float), _ta.HT_SINE, ret_idx=ret_idx))


def HT_TRENDMODE(close: Expr) -> Expr:  # ['integer']
    return close.map_batches(lambda x1: batches_i1_o1(x1.to_numpy().astype(float), _ta.HT_TRENDMODE))


def ADD(high: Expr, low: Expr) -> Expr:  # ['real']
    return map_batches([high, low], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.ADD))


def DIV(high: Expr, low: Expr) -> Expr:  # ['real']
    return map_batches([high, low], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.DIV))


def MAX(close: Expr, timeperiod: int = 30) -> Expr:  # ['real']
    return close.map_batches(lambda x1: batches_i1_o1(x1.to_numpy().astype(float), _ta.MAX, timeperiod))


def MAXINDEX(close: Expr, timeperiod: int = 30) -> Expr:  # ['integer']
    return close.map_batches(lambda x1: batches_i1_o1(x1.to_numpy().astype(float), _ta.MAXINDEX, timeperiod))


def MIN(close: Expr, timeperiod: int = 30) -> Expr:  # ['real']
    return close.map_batches(lambda x1: batches_i1_o1(x1.to_numpy().astype(float), _ta.MIN, timeperiod))


def MININDEX(close: Expr, timeperiod: int = 30) -> Expr:  # ['integer']
    return close.map_batches(lambda x1: batches_i1_o1(x1.to_numpy().astype(float), _ta.MININDEX, timeperiod))


def MINMAX(close: Expr, timeperiod: int = 30, ret_idx: int = 1) -> Expr:  # ['min', 'max']
    return close.map_batches(lambda x1: batches_i1_o2(x1.to_numpy().astype(float), _ta.MINMAX, timeperiod, ret_idx=ret_idx))


def MINMAXINDEX(close: Expr, timeperiod: int = 30, ret_idx: int = 1) -> Expr:  # ['minidx', 'maxidx']
    return close.map_batches(lambda x1: batches_i1_o2(x1.to_numpy().astype(float), _ta.MINMAXINDEX, timeperiod, ret_idx=ret_idx))


def MULT(high: Expr, low: Expr) -> Expr:  # ['real']
    return map_batches([high, low], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.MULT))


def SUB(high: Expr, low: Expr) -> Expr:  # ['real']
    return map_batches([high, low], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.SUB))


def SUM(close: Expr, timeperiod: int = 30) -> Expr:  # ['real']
    return close.map_batches(lambda x1: batches_i1_o1(x1.to_numpy().astype(float), _ta.SUM, timeperiod))


def ACOS(close: Expr) -> Expr:  # ['real']
    return close.map_batches(lambda x1: batches_i1_o1(x1.to_numpy().astype(float), _ta.ACOS))


def ASIN(close: Expr) -> Expr:  # ['real']
    return close.map_batches(lambda x1: batches_i1_o1(x1.to_numpy().astype(float), _ta.ASIN))


def ATAN(close: Expr) -> Expr:  # ['real']
    return close.map_batches(lambda x1: batches_i1_o1(x1.to_numpy().astype(float), _ta.ATAN))


def CEIL(close: Expr) -> Expr:  # ['real']
    return close.map_batches(lambda x1: batches_i1_o1(x1.to_numpy().astype(float), _ta.CEIL))


def COS(close: Expr) -> Expr:  # ['real']
    return close.map_batches(lambda x1: batches_i1_o1(x1.to_numpy().astype(float), _ta.COS))


def COSH(close: Expr) -> Expr:  # ['real']
    return close.map_batches(lambda x1: batches_i1_o1(x1.to_numpy().astype(float), _ta.COSH))


def EXP(close: Expr) -> Expr:  # ['real']
    return close.map_batches(lambda x1: batches_i1_o1(x1.to_numpy().astype(float), _ta.EXP))


def FLOOR(close: Expr) -> Expr:  # ['real']
    return close.map_batches(lambda x1: batches_i1_o1(x1.to_numpy().astype(float), _ta.FLOOR))


def LN(close: Expr) -> Expr:  # ['real']
    return close.map_batches(lambda x1: batches_i1_o1(x1.to_numpy().astype(float), _ta.LN))


def LOG10(close: Expr) -> Expr:  # ['real']
    return close.map_batches(lambda x1: batches_i1_o1(x1.to_numpy().astype(float), _ta.LOG10))


def SIN(close: Expr) -> Expr:  # ['real']
    return close.map_batches(lambda x1: batches_i1_o1(x1.to_numpy().astype(float), _ta.SIN))


def SINH(close: Expr) -> Expr:  # ['real']
    return close.map_batches(lambda x1: batches_i1_o1(x1.to_numpy().astype(float), _ta.SINH))


def SQRT(close: Expr) -> Expr:  # ['real']
    return close.map_batches(lambda x1: batches_i1_o1(x1.to_numpy().astype(float), _ta.SQRT))


def TAN(close: Expr) -> Expr:  # ['real']
    return close.map_batches(lambda x1: batches_i1_o1(x1.to_numpy().astype(float), _ta.TAN))


def TANH(close: Expr) -> Expr:  # ['real']
    return close.map_batches(lambda x1: batches_i1_o1(x1.to_numpy().astype(float), _ta.TANH))


def ADX(high: Expr, low: Expr, close: Expr, timeperiod: int = 14) -> Expr:  # ['real']
    return map_batches([high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.ADX, timeperiod))


def ADXR(high: Expr, low: Expr, close: Expr, timeperiod: int = 14) -> Expr:  # ['real']
    return map_batches([high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.ADXR, timeperiod))


def APO(close: Expr, fastperiod: int = 12, slowperiod: int = 26, matype: int = 0) -> Expr:  # ['real']
    return close.map_batches(lambda x1: batches_i1_o1(x1.to_numpy().astype(float), _ta.APO, fastperiod, slowperiod, matype))


def AROON(high: Expr, low: Expr, timeperiod: int = 14, ret_idx: int = 1) -> Expr:  # ['aroondown', 'aroonup']
    return map_batches([high, low], lambda xx: batches_i2_o2([x1.to_numpy().astype(float) for x1 in xx], _ta.AROON, timeperiod, ret_idx=ret_idx))


def AROONOSC(high: Expr, low: Expr, timeperiod: int = 14) -> Expr:  # ['real']
    return map_batches([high, low], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.AROONOSC, timeperiod))


def BOP(open: Expr, high: Expr, low: Expr, close: Expr) -> Expr:  # ['real']
    return map_batches([open, high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.BOP))


def CCI(high: Expr, low: Expr, close: Expr, timeperiod: int = 14) -> Expr:  # ['real']
    return map_batches([high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.CCI, timeperiod))


def CMO(close: Expr, timeperiod: int = 14) -> Expr:  # ['real']
    return close.map_batches(lambda x1: batches_i1_o1(x1.to_numpy().astype(float), _ta.CMO, timeperiod))


def DX(high: Expr, low: Expr, close: Expr, timeperiod: int = 14) -> Expr:  # ['real']
    return map_batches([high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.DX, timeperiod))


def MACD(close: Expr, fastperiod: int = 12, slowperiod: int = 26, signalperiod: int = 9, ret_idx: int = 2) -> Expr:  # ['macd', 'macdsignal', 'macdhist']
    return close.map_batches(lambda x1: batches_i1_o2(x1.to_numpy().astype(float), _ta.MACD, fastperiod, slowperiod, signalperiod, ret_idx=ret_idx))


def MACDEXT(close: Expr, fastperiod: int = 12, fastmatype: int = 0, slowperiod: int = 26, slowmatype: int = 0, signalperiod: int = 9, signalmatype: int = 0, ret_idx: int = 2) -> Expr:  # ['macd', 'macdsignal', 'macdhist']
    return close.map_batches(lambda x1: batches_i1_o2(x1.to_numpy().astype(float), _ta.MACDEXT, fastperiod, fastmatype, slowperiod, slowmatype, signalperiod, signalmatype, ret_idx=ret_idx))


def MACDFIX(close: Expr, signalperiod: int = 9, ret_idx: int = 2) -> Expr:  # ['macd', 'macdsignal', 'macdhist']
    return close.map_batches(lambda x1: batches_i1_o2(x1.to_numpy().astype(float), _ta.MACDFIX, signalperiod, ret_idx=ret_idx))


def MFI(high: Expr, low: Expr, close: Expr, volume: Expr, timeperiod: int = 14) -> Expr:  # ['real']
    return map_batches([high, low, close, volume], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.MFI, timeperiod))


def MINUS_DI(high: Expr, low: Expr, close: Expr, timeperiod: int = 14) -> Expr:  # ['real']
    return map_batches([high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.MINUS_DI, timeperiod))


def MINUS_DM(high: Expr, low: Expr, timeperiod: int = 14) -> Expr:  # ['real']
    return map_batches([high, low], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.MINUS_DM, timeperiod))


def MOM(close: Expr, timeperiod: int = 10) -> Expr:  # ['real']
    return close.map_batches(lambda x1: batches_i1_o1(x1.to_numpy().astype(float), _ta.MOM, timeperiod))


def PLUS_DI(high: Expr, low: Expr, close: Expr, timeperiod: int = 14) -> Expr:  # ['real']
    return map_batches([high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.PLUS_DI, timeperiod))


def PLUS_DM(high: Expr, low: Expr, timeperiod: int = 14) -> Expr:  # ['real']
    return map_batches([high, low], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.PLUS_DM, timeperiod))


def PPO(close: Expr, fastperiod: int = 12, slowperiod: int = 26, matype: int = 0) -> Expr:  # ['real']
    return close.map_batches(lambda x1: batches_i1_o1(x1.to_numpy().astype(float), _ta.PPO, fastperiod, slowperiod, matype))


def ROC(close: Expr, timeperiod: int = 10) -> Expr:  # ['real']
    return close.map_batches(lambda x1: batches_i1_o1(x1.to_numpy().astype(float), _ta.ROC, timeperiod))


def ROCP(close: Expr, timeperiod: int = 10) -> Expr:  # ['real']
    return close.map_batches(lambda x1: batches_i1_o1(x1.to_numpy().astype(float), _ta.ROCP, timeperiod))


def ROCR(close: Expr, timeperiod: int = 10) -> Expr:  # ['real']
    return close.map_batches(lambda x1: batches_i1_o1(x1.to_numpy().astype(float), _ta.ROCR, timeperiod))


def ROCR100(close: Expr, timeperiod: int = 10) -> Expr:  # ['real']
    return close.map_batches(lambda x1: batches_i1_o1(x1.to_numpy().astype(float), _ta.ROCR100, timeperiod))


def RSI(close: Expr, timeperiod: int = 14) -> Expr:  # ['real']
    return close.map_batches(lambda x1: batches_i1_o1(x1.to_numpy().astype(float), _ta.RSI, timeperiod))


def STOCH(high: Expr, low: Expr, close: Expr, fastk_period: int = 5, slowk_period: int = 3, slowk_matype: int = 0, slowd_period: int = 3, slowd_matype: int = 0, ret_idx: int = 1) -> Expr:  # ['slowk', 'slowd']
    return map_batches([high, low, close], lambda xx: batches_i2_o2([x1.to_numpy().astype(float) for x1 in xx], _ta.STOCH, fastk_period, slowk_period, slowk_matype, slowd_period, slowd_matype, ret_idx=ret_idx))


def STOCHF(high: Expr, low: Expr, close: Expr, fastk_period: int = 5, fastd_period: int = 3, fastd_matype: int = 0, ret_idx: int = 1) -> Expr:  # ['fastk', 'fastd']
    return map_batches([high, low, close], lambda xx: batches_i2_o2([x1.to_numpy().astype(float) for x1 in xx], _ta.STOCHF, fastk_period, fastd_period, fastd_matype, ret_idx=ret_idx))


def STOCHRSI(close: Expr, timeperiod: int = 14, fastk_period: int = 5, fastd_period: int = 3, fastd_matype: int = 0, ret_idx: int = 1) -> Expr:  # ['fastk', 'fastd']
    return close.map_batches(lambda x1: batches_i1_o2(x1.to_numpy().astype(float), _ta.STOCHRSI, timeperiod, fastk_period, fastd_period, fastd_matype, ret_idx=ret_idx))


def TRIX(close: Expr, timeperiod: int = 30) -> Expr:  # ['real']
    return close.map_batches(lambda x1: batches_i1_o1(x1.to_numpy().astype(float), _ta.TRIX, timeperiod))


def ULTOSC(high: Expr, low: Expr, close: Expr, timeperiod1: int = 7, timeperiod2: int = 14, timeperiod3: int = 28) -> Expr:  # ['real']
    return map_batches([high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.ULTOSC, timeperiod1, timeperiod2, timeperiod3))


def WILLR(high: Expr, low: Expr, close: Expr, timeperiod: int = 14) -> Expr:  # ['real']
    return map_batches([high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.WILLR, timeperiod))


def BBANDS(close: Expr, timeperiod: int = 5, nbdevup: float = 2.0, nbdevdn: float = 2.0, matype: int = 0, ret_idx: int = 2) -> Expr:  # ['upperband', 'middleband', 'lowerband']
    return close.map_batches(lambda x1: batches_i1_o2(x1.to_numpy().astype(float), _ta.BBANDS, timeperiod, nbdevup, nbdevdn, matype, ret_idx=ret_idx))


def DEMA(close: Expr, timeperiod: int = 30) -> Expr:  # ['real']
    return close.map_batches(lambda x1: batches_i1_o1(x1.to_numpy().astype(float), _ta.DEMA, timeperiod))


def EMA(close: Expr, timeperiod: int = 30) -> Expr:  # ['real']
    return close.map_batches(lambda x1: batches_i1_o1(x1.to_numpy().astype(float), _ta.EMA, timeperiod))


def HT_TRENDLINE(close: Expr) -> Expr:  # ['real']
    return close.map_batches(lambda x1: batches_i1_o1(x1.to_numpy().astype(float), _ta.HT_TRENDLINE))


def KAMA(close: Expr, timeperiod: int = 30) -> Expr:  # ['real']
    return close.map_batches(lambda x1: batches_i1_o1(x1.to_numpy().astype(float), _ta.KAMA, timeperiod))


def MA(close: Expr, timeperiod: int = 30, matype: int = 0) -> Expr:  # ['real']
    return close.map_batches(lambda x1: batches_i1_o1(x1.to_numpy().astype(float), _ta.MA, timeperiod, matype))


def MAMA(close: Expr, fastlimit: float = 0.5, slowlimit: float = 0.05, ret_idx: int = 1) -> Expr:  # ['mama', 'fama']
    return close.map_batches(lambda x1: batches_i1_o2(x1.to_numpy().astype(float), _ta.MAMA, fastlimit, slowlimit, ret_idx=ret_idx))


def MAVP(close: Expr, periods: Expr, minperiod: int = 2, maxperiod: int = 30, matype: int = 0) -> Expr:  # ['real']
    return map_batches([close, periods], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.MAVP, minperiod, maxperiod, matype))


def MIDPOINT(close: Expr, timeperiod: int = 14) -> Expr:  # ['real']
    return close.map_batches(lambda x1: batches_i1_o1(x1.to_numpy().astype(float), _ta.MIDPOINT, timeperiod))


def MIDPRICE(high: Expr, low: Expr, timeperiod: int = 14) -> Expr:  # ['real']
    return map_batches([high, low], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.MIDPRICE, timeperiod))


def SAR(high: Expr, low: Expr, acceleration: float = 0.02, maximum: float = 0.2) -> Expr:  # ['real']
    return map_batches([high, low], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.SAR, acceleration, maximum))


def SAREXT(high: Expr, low: Expr, startvalue: float = 0.0, offsetonreverse: float = 0.0, accelerationinitlong: float = 0.02, accelerationlong: float = 0.02, accelerationmaxlong: float = 0.2, accelerationinitshort: float = 0.02, accelerationshort: float = 0.02, accelerationmaxshort: float = 0.2) -> Expr:  # ['real']
    return map_batches([high, low], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.SAREXT, startvalue, offsetonreverse, accelerationinitlong, accelerationlong, accelerationmaxlong, accelerationinitshort, accelerationshort, accelerationmaxshort))


def SMA(close: Expr, timeperiod: int = 30) -> Expr:  # ['real']
    return close.map_batches(lambda x1: batches_i1_o1(x1.to_numpy().astype(float), _ta.SMA, timeperiod))


def T3(close: Expr, timeperiod: int = 5, vfactor: float = 0.7) -> Expr:  # ['real']
    return close.map_batches(lambda x1: batches_i1_o1(x1.to_numpy().astype(float), _ta.T3, timeperiod, vfactor))


def TEMA(close: Expr, timeperiod: int = 30) -> Expr:  # ['real']
    return close.map_batches(lambda x1: batches_i1_o1(x1.to_numpy().astype(float), _ta.TEMA, timeperiod))


def TRIMA(close: Expr, timeperiod: int = 30) -> Expr:  # ['real']
    return close.map_batches(lambda x1: batches_i1_o1(x1.to_numpy().astype(float), _ta.TRIMA, timeperiod))


def WMA(close: Expr, timeperiod: int = 30) -> Expr:  # ['real']
    return close.map_batches(lambda x1: batches_i1_o1(x1.to_numpy().astype(float), _ta.WMA, timeperiod))


def CDL2CROWS(open: Expr, high: Expr, low: Expr, close: Expr) -> Expr:  # ['integer']
    return map_batches([open, high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.CDL2CROWS))


def CDL3BLACKCROWS(open: Expr, high: Expr, low: Expr, close: Expr) -> Expr:  # ['integer']
    return map_batches([open, high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.CDL3BLACKCROWS))


def CDL3INSIDE(open: Expr, high: Expr, low: Expr, close: Expr) -> Expr:  # ['integer']
    return map_batches([open, high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.CDL3INSIDE))


def CDL3LINESTRIKE(open: Expr, high: Expr, low: Expr, close: Expr) -> Expr:  # ['integer']
    return map_batches([open, high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.CDL3LINESTRIKE))


def CDL3OUTSIDE(open: Expr, high: Expr, low: Expr, close: Expr) -> Expr:  # ['integer']
    return map_batches([open, high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.CDL3OUTSIDE))


def CDL3STARSINSOUTH(open: Expr, high: Expr, low: Expr, close: Expr) -> Expr:  # ['integer']
    return map_batches([open, high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.CDL3STARSINSOUTH))


def CDL3WHITESOLDIERS(open: Expr, high: Expr, low: Expr, close: Expr) -> Expr:  # ['integer']
    return map_batches([open, high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.CDL3WHITESOLDIERS))


def CDLABANDONEDBABY(open: Expr, high: Expr, low: Expr, close: Expr, penetration: float = 0.3) -> Expr:  # ['integer']
    return map_batches([open, high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.CDLABANDONEDBABY, penetration))


def CDLADVANCEBLOCK(open: Expr, high: Expr, low: Expr, close: Expr) -> Expr:  # ['integer']
    return map_batches([open, high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.CDLADVANCEBLOCK))


def CDLBELTHOLD(open: Expr, high: Expr, low: Expr, close: Expr) -> Expr:  # ['integer']
    return map_batches([open, high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.CDLBELTHOLD))


def CDLBREAKAWAY(open: Expr, high: Expr, low: Expr, close: Expr) -> Expr:  # ['integer']
    return map_batches([open, high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.CDLBREAKAWAY))


def CDLCLOSINGMARUBOZU(open: Expr, high: Expr, low: Expr, close: Expr) -> Expr:  # ['integer']
    return map_batches([open, high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.CDLCLOSINGMARUBOZU))


def CDLCONCEALBABYSWALL(open: Expr, high: Expr, low: Expr, close: Expr) -> Expr:  # ['integer']
    return map_batches([open, high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.CDLCONCEALBABYSWALL))


def CDLCOUNTERATTACK(open: Expr, high: Expr, low: Expr, close: Expr) -> Expr:  # ['integer']
    return map_batches([open, high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.CDLCOUNTERATTACK))


def CDLDARKCLOUDCOVER(open: Expr, high: Expr, low: Expr, close: Expr, penetration: float = 0.5) -> Expr:  # ['integer']
    return map_batches([open, high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.CDLDARKCLOUDCOVER, penetration))


def CDLDOJI(open: Expr, high: Expr, low: Expr, close: Expr) -> Expr:  # ['integer']
    return map_batches([open, high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.CDLDOJI))


def CDLDOJISTAR(open: Expr, high: Expr, low: Expr, close: Expr) -> Expr:  # ['integer']
    return map_batches([open, high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.CDLDOJISTAR))


def CDLDRAGONFLYDOJI(open: Expr, high: Expr, low: Expr, close: Expr) -> Expr:  # ['integer']
    return map_batches([open, high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.CDLDRAGONFLYDOJI))


def CDLENGULFING(open: Expr, high: Expr, low: Expr, close: Expr) -> Expr:  # ['integer']
    return map_batches([open, high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.CDLENGULFING))


def CDLEVENINGDOJISTAR(open: Expr, high: Expr, low: Expr, close: Expr, penetration: float = 0.3) -> Expr:  # ['integer']
    return map_batches([open, high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.CDLEVENINGDOJISTAR, penetration))


def CDLEVENINGSTAR(open: Expr, high: Expr, low: Expr, close: Expr, penetration: float = 0.3) -> Expr:  # ['integer']
    return map_batches([open, high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.CDLEVENINGSTAR, penetration))


def CDLGAPSIDESIDEWHITE(open: Expr, high: Expr, low: Expr, close: Expr) -> Expr:  # ['integer']
    return map_batches([open, high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.CDLGAPSIDESIDEWHITE))


def CDLGRAVESTONEDOJI(open: Expr, high: Expr, low: Expr, close: Expr) -> Expr:  # ['integer']
    return map_batches([open, high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.CDLGRAVESTONEDOJI))


def CDLHAMMER(open: Expr, high: Expr, low: Expr, close: Expr) -> Expr:  # ['integer']
    return map_batches([open, high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.CDLHAMMER))


def CDLHANGINGMAN(open: Expr, high: Expr, low: Expr, close: Expr) -> Expr:  # ['integer']
    return map_batches([open, high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.CDLHANGINGMAN))


def CDLHARAMI(open: Expr, high: Expr, low: Expr, close: Expr) -> Expr:  # ['integer']
    return map_batches([open, high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.CDLHARAMI))


def CDLHARAMICROSS(open: Expr, high: Expr, low: Expr, close: Expr) -> Expr:  # ['integer']
    return map_batches([open, high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.CDLHARAMICROSS))


def CDLHIGHWAVE(open: Expr, high: Expr, low: Expr, close: Expr) -> Expr:  # ['integer']
    return map_batches([open, high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.CDLHIGHWAVE))


def CDLHIKKAKE(open: Expr, high: Expr, low: Expr, close: Expr) -> Expr:  # ['integer']
    return map_batches([open, high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.CDLHIKKAKE))


def CDLHIKKAKEMOD(open: Expr, high: Expr, low: Expr, close: Expr) -> Expr:  # ['integer']
    return map_batches([open, high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.CDLHIKKAKEMOD))


def CDLHOMINGPIGEON(open: Expr, high: Expr, low: Expr, close: Expr) -> Expr:  # ['integer']
    return map_batches([open, high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.CDLHOMINGPIGEON))


def CDLIDENTICAL3CROWS(open: Expr, high: Expr, low: Expr, close: Expr) -> Expr:  # ['integer']
    return map_batches([open, high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.CDLIDENTICAL3CROWS))


def CDLINNECK(open: Expr, high: Expr, low: Expr, close: Expr) -> Expr:  # ['integer']
    return map_batches([open, high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.CDLINNECK))


def CDLINVERTEDHAMMER(open: Expr, high: Expr, low: Expr, close: Expr) -> Expr:  # ['integer']
    return map_batches([open, high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.CDLINVERTEDHAMMER))


def CDLKICKING(open: Expr, high: Expr, low: Expr, close: Expr) -> Expr:  # ['integer']
    return map_batches([open, high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.CDLKICKING))


def CDLKICKINGBYLENGTH(open: Expr, high: Expr, low: Expr, close: Expr) -> Expr:  # ['integer']
    return map_batches([open, high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.CDLKICKINGBYLENGTH))


def CDLLADDERBOTTOM(open: Expr, high: Expr, low: Expr, close: Expr) -> Expr:  # ['integer']
    return map_batches([open, high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.CDLLADDERBOTTOM))


def CDLLONGLEGGEDDOJI(open: Expr, high: Expr, low: Expr, close: Expr) -> Expr:  # ['integer']
    return map_batches([open, high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.CDLLONGLEGGEDDOJI))


def CDLLONGLINE(open: Expr, high: Expr, low: Expr, close: Expr) -> Expr:  # ['integer']
    return map_batches([open, high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.CDLLONGLINE))


def CDLMARUBOZU(open: Expr, high: Expr, low: Expr, close: Expr) -> Expr:  # ['integer']
    return map_batches([open, high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.CDLMARUBOZU))


def CDLMATCHINGLOW(open: Expr, high: Expr, low: Expr, close: Expr) -> Expr:  # ['integer']
    return map_batches([open, high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.CDLMATCHINGLOW))


def CDLMATHOLD(open: Expr, high: Expr, low: Expr, close: Expr, penetration: float = 0.5) -> Expr:  # ['integer']
    return map_batches([open, high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.CDLMATHOLD, penetration))


def CDLMORNINGDOJISTAR(open: Expr, high: Expr, low: Expr, close: Expr, penetration: float = 0.3) -> Expr:  # ['integer']
    return map_batches([open, high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.CDLMORNINGDOJISTAR, penetration))


def CDLMORNINGSTAR(open: Expr, high: Expr, low: Expr, close: Expr, penetration: float = 0.3) -> Expr:  # ['integer']
    return map_batches([open, high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.CDLMORNINGSTAR, penetration))


def CDLONNECK(open: Expr, high: Expr, low: Expr, close: Expr) -> Expr:  # ['integer']
    return map_batches([open, high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.CDLONNECK))


def CDLPIERCING(open: Expr, high: Expr, low: Expr, close: Expr) -> Expr:  # ['integer']
    return map_batches([open, high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.CDLPIERCING))


def CDLRICKSHAWMAN(open: Expr, high: Expr, low: Expr, close: Expr) -> Expr:  # ['integer']
    return map_batches([open, high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.CDLRICKSHAWMAN))


def CDLRISEFALL3METHODS(open: Expr, high: Expr, low: Expr, close: Expr) -> Expr:  # ['integer']
    return map_batches([open, high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.CDLRISEFALL3METHODS))


def CDLSEPARATINGLINES(open: Expr, high: Expr, low: Expr, close: Expr) -> Expr:  # ['integer']
    return map_batches([open, high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.CDLSEPARATINGLINES))


def CDLSHOOTINGSTAR(open: Expr, high: Expr, low: Expr, close: Expr) -> Expr:  # ['integer']
    return map_batches([open, high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.CDLSHOOTINGSTAR))


def CDLSHORTLINE(open: Expr, high: Expr, low: Expr, close: Expr) -> Expr:  # ['integer']
    return map_batches([open, high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.CDLSHORTLINE))


def CDLSPINNINGTOP(open: Expr, high: Expr, low: Expr, close: Expr) -> Expr:  # ['integer']
    return map_batches([open, high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.CDLSPINNINGTOP))


def CDLSTALLEDPATTERN(open: Expr, high: Expr, low: Expr, close: Expr) -> Expr:  # ['integer']
    return map_batches([open, high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.CDLSTALLEDPATTERN))


def CDLSTICKSANDWICH(open: Expr, high: Expr, low: Expr, close: Expr) -> Expr:  # ['integer']
    return map_batches([open, high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.CDLSTICKSANDWICH))


def CDLTAKURI(open: Expr, high: Expr, low: Expr, close: Expr) -> Expr:  # ['integer']
    return map_batches([open, high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.CDLTAKURI))


def CDLTASUKIGAP(open: Expr, high: Expr, low: Expr, close: Expr) -> Expr:  # ['integer']
    return map_batches([open, high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.CDLTASUKIGAP))


def CDLTHRUSTING(open: Expr, high: Expr, low: Expr, close: Expr) -> Expr:  # ['integer']
    return map_batches([open, high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.CDLTHRUSTING))


def CDLTRISTAR(open: Expr, high: Expr, low: Expr, close: Expr) -> Expr:  # ['integer']
    return map_batches([open, high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.CDLTRISTAR))


def CDLUNIQUE3RIVER(open: Expr, high: Expr, low: Expr, close: Expr) -> Expr:  # ['integer']
    return map_batches([open, high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.CDLUNIQUE3RIVER))


def CDLUPSIDEGAP2CROWS(open: Expr, high: Expr, low: Expr, close: Expr) -> Expr:  # ['integer']
    return map_batches([open, high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.CDLUPSIDEGAP2CROWS))


def CDLXSIDEGAP3METHODS(open: Expr, high: Expr, low: Expr, close: Expr) -> Expr:  # ['integer']
    return map_batches([open, high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.CDLXSIDEGAP3METHODS))


def AVGPRICE(open: Expr, high: Expr, low: Expr, close: Expr) -> Expr:  # ['real']
    return map_batches([open, high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.AVGPRICE))


def MEDPRICE(high: Expr, low: Expr) -> Expr:  # ['real']
    return map_batches([high, low], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.MEDPRICE))


def TYPPRICE(high: Expr, low: Expr, close: Expr) -> Expr:  # ['real']
    return map_batches([high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.TYPPRICE))


def WCLPRICE(high: Expr, low: Expr, close: Expr) -> Expr:  # ['real']
    return map_batches([high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.WCLPRICE))


def BETA(high: Expr, low: Expr, timeperiod: int = 5) -> Expr:  # ['real']
    return map_batches([high, low], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.BETA, timeperiod))


def CORREL(high: Expr, low: Expr, timeperiod: int = 30) -> Expr:  # ['real']
    return map_batches([high, low], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.CORREL, timeperiod))


def LINEARREG(close: Expr, timeperiod: int = 14) -> Expr:  # ['real']
    return close.map_batches(lambda x1: batches_i1_o1(x1.to_numpy().astype(float), _ta.LINEARREG, timeperiod))


def LINEARREG_ANGLE(close: Expr, timeperiod: int = 14) -> Expr:  # ['real']
    return close.map_batches(lambda x1: batches_i1_o1(x1.to_numpy().astype(float), _ta.LINEARREG_ANGLE, timeperiod))


def LINEARREG_INTERCEPT(close: Expr, timeperiod: int = 14) -> Expr:  # ['real']
    return close.map_batches(lambda x1: batches_i1_o1(x1.to_numpy().astype(float), _ta.LINEARREG_INTERCEPT, timeperiod))


def LINEARREG_SLOPE(close: Expr, timeperiod: int = 14) -> Expr:  # ['real']
    return close.map_batches(lambda x1: batches_i1_o1(x1.to_numpy().astype(float), _ta.LINEARREG_SLOPE, timeperiod))


def STDDEV(close: Expr, timeperiod: int = 5, nbdev: float = 1.0) -> Expr:  # ['real']
    return close.map_batches(lambda x1: batches_i1_o1(x1.to_numpy().astype(float), _ta.STDDEV, timeperiod, nbdev))


def TSF(close: Expr, timeperiod: int = 14) -> Expr:  # ['real']
    return close.map_batches(lambda x1: batches_i1_o1(x1.to_numpy().astype(float), _ta.TSF, timeperiod))


def VAR(close: Expr, timeperiod: int = 5, nbdev: float = 1.0) -> Expr:  # ['real']
    return close.map_batches(lambda x1: batches_i1_o1(x1.to_numpy().astype(float), _ta.VAR, timeperiod, nbdev))


def ATR(high: Expr, low: Expr, close: Expr, timeperiod: int = 14) -> Expr:  # ['real']
    return map_batches([high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.ATR, timeperiod))


def NATR(high: Expr, low: Expr, close: Expr, timeperiod: int = 14) -> Expr:  # ['real']
    return map_batches([high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.NATR, timeperiod))


def TRANGE(high: Expr, low: Expr, close: Expr) -> Expr:  # ['real']
    return map_batches([high, low, close], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.TRANGE))


def AD(high: Expr, low: Expr, close: Expr, volume: Expr) -> Expr:  # ['real']
    return map_batches([high, low, close, volume], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.AD))


def ADOSC(high: Expr, low: Expr, close: Expr, volume: Expr, fastperiod: int = 3, slowperiod: int = 10) -> Expr:  # ['real']
    return map_batches([high, low, close, volume], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.ADOSC, fastperiod, slowperiod))


def OBV(close: Expr, volume: Expr) -> Expr:  # ['real']
    return map_batches([close, volume], lambda xx: batches_i2_o1([x1.to_numpy().astype(float) for x1 in xx], _ta.OBV))



"""
The following comes from Trading Systems and Methods, Chapter 1, Measuring Noise

We do not provide ATR and STD here.


ATR与STD也是一种度量波动的方法，这里不再提供

以下方法来自于Trading Systems and Methods, Chapter 1, Measuring Noise

References
----------
https://zhuanlan.zhihu.com/p/544744582

"""
import numpy as np
from polars import Expr


def ts_efficiency_ratio(close: Expr, timeperiod: int = 14) -> Expr:
    """效率系数。值越大，噪音越小。最大值为1，最小值为0

    本质上是位移除以路程
    """
    t1 = close.diff(timeperiod).abs()
    t2 = close.diff(1).abs().rolling_sum(timeperiod)
    return t1 / t2


def ts_price_density(high: Expr, low: Expr, timeperiod: int = 14) -> Expr:
    """价格密度。值越大，噪音越大

    如果K线高低相连，上涨为1，下跌也为1
    如果K线高低平行，值大于1，最大为timeperiod
    """
    t1 = (high - low).rolling_sum(timeperiod)
    t2 = high.rolling_max(timeperiod) - low.rolling_min(timeperiod)
    return t1 / t2


def ts_fractal_dimension(high: Expr, low: Expr, close: Expr, timeperiod: int = 14) -> Expr:
    """分形维度。值越大，噪音越大"""
    n1 = (1 / timeperiod) ** 2
    n2 = np.log(2)
    n3 = np.log(2 * timeperiod)

    t1 = high.rolling_max(timeperiod) - low.rolling_min(timeperiod)
    t2 = close.diff(1)

    L = (n1 + t2 / t1).sqrt().rolling_sum(timeperiod)
    return 1 + (L.log() + n2) / n3



"""
Algorithm in this file is slow, and has been replaced by numba and other methods
本目录下算法有些慢，已经用numba等其它方法代替
"""
import numpy as np
from polars import Series, Expr


def _arg_max(x: Series):
    """
    Notes
    -----
    TODO 等polars推出rolling_arg_max(reverse=True)这个问题能好转

    """
    # return x[::-1].arg_max()
    # return x.reverse().arg_max() # 正确，但太慢
    return len(x) - 1 - x.arg_max()  # 有多个最大值相同时，靠前的值会被记录下来，导致结果偏大


def ts_arg_max(x: Expr, d: int = 5) -> Expr:
    # WorldQuant中最大值为今天返回0，为昨天返回1
    return x.rolling_map(_arg_max, d)


def _arg_min(x: Series):
    return len(x) - 1 - x.arg_min()


def ts_arg_min(x: Expr, d: int = 5) -> Expr:
    return x.rolling_map(_arg_min, d)


def ts_product(x: Expr, d: int = 5) -> Expr:
    return x.rolling_map(np.nanprod, d)



import pandas as pd
from polars import Expr, when, Series, Int16, max_horizontal


def _qcut(x1: Series, q: int) -> Series:
    # TODO 等待提供
    if x1.null_count() == len(x1):
        return x1
    else:
        return pd.qcut(x1, q, labels=False, duplicates='drop')


def _rank_qcut(x: Expr, q: int = 10) -> Expr:
    """结果与qcut基本一样，速度快三倍"""
    a = x.rank(method='min') - 1.001
    b = max_horizontal(x.count() - 1, 1)
    return (a / b * q).cast(Int16)


def cs_bucket(x: Expr, q: int = 10) -> Expr:
    """Convert float values into indexes for user-specified buckets. Bucket is useful for creating group values, which can be passed to group operators as input."""
    # TODO 等官方提供原生功能
    # return x.map_batches(lambda x1: Series(_qcut(x1, q), nan_to_null=True, dtype=Int16))
    return _rank_qcut(x, q)


def clamp(x: Expr, lower: float = 0, upper: float = 0, inverse: bool = False, mask: float = None) -> Expr:
    """Limits input value between lower and upper bound in inverse = false mode (which is default). Alternatively, when inverse = true, values between bounds are replaced with mask, while values outside bounds are left as is."""
    if inverse:
        # mask is one of: 'nearest_bound', 'mean', 'NAN' or any floating point number
        return when((x < lower) | (x > upper)).then(x).otherwise(mask)
    else:
        return x.clip(lower, upper)


def filter_(x: Expr, h: str = "1, 2, 3, 4", t: str = "0.5") -> Expr:
    """Used to filter the value and allows to create filters like linear or exponential decay."""
    raise


def keep(x: Expr, f: float, period: int = 5) -> Expr:
    """This operator outputs value x when f changes and continues to do that for “period” days after f stopped changing. After “period” days since last change of f, NaN is output."""
    raise


def left_tail(x: Expr, maximum: float = 0) -> Expr:
    """NaN everything greater than maximum, maximum should be constant."""
    return when(x <= maximum).then(x).otherwise(None)


def pasteurize(x: Expr) -> Expr:
    """Set to NaN if x is INF or if the underlying instrument is not in the Alpha universe"""
    # TODO: 不在票池中的的功能无法表示
    # TODO: 与purify好像没啥区别
    return when(x.is_finite()).then(x).otherwise(None)


def purify(x: Expr) -> Expr:
    """Clear infinities (+inf, -inf) by replacing with NaN."""
    return when(x.is_finite()).then(x).otherwise(None)


def fill_nan(x: Expr) -> Expr:
    """fill nan by null
    填充nan为null"""
    return x.fill_nan(None)


def fill_zero(x: Expr) -> Expr:
    """填充null为0"""
    return x.fill_null(0)


def right_tail(x: Expr, minimum: float = 0) -> Expr:
    """NaN everything less than minimum, minimum should be constant."""
    return when(x >= minimum).then(x).otherwise(None)


def sigmoid(x: Expr) -> Expr:
    """Returns 1 / (1 + exp(-x))"""
    return 1 / (1 + (-x).exp())


def tail(x: Expr, lower: float = 0, upper: float = 0, newval: float = 0) -> Expr:
    """If (x > lower AND x < upper) return newval, else return x. Lower, upper, newval should be constants. """
    # TODO 与clamp一样?
    return when((x <= lower) | (x >= upper)).then(x).otherwise(newval)



from math import ceil, floor

from polars import Expr

from polars_ta.ta.operators import MAX
from polars_ta.ta.operators import MIN
from polars_ta.ta.statistic import STDDEV
from polars_ta.wq.time_series import ts_decay_linear as WMA  # noqa
from polars_ta.wq.time_series import ts_mean as SMA  # noqa


def BBANDS_upperband(close: Expr, timeperiod: int = 5, nbdevup: float = 2) -> Expr:
    """Bollinger Bands Upper Band
    布林线上轨

    Notes
    -----
    1. You may create a new functino based on this, to replace the middle band
    2. use negative values in `nbdevup` for the lower band

    1. 想替换中线算法时，参考此代码新建一个函数
    2. 想生成下轨时，nbdevup用负数
    """
    return SMA(close, timeperiod) + STDDEV(close, timeperiod, nbdevup)


def DEMA(close: Expr, timeperiod: int = 30) -> Expr:
    EMA1 = EMA(close, timeperiod)
    EMA2 = EMA(EMA1, timeperiod)
    return EMA1 * 2 - EMA2


def EMA(close: Expr, timeperiod: int = 30) -> Expr:
    """

    References
    ----------
    https://pola-rs.github.io/polars/py-polars/html/reference/expressions/api/polars.Expr.ewm_mean.html#polars.Expr.ewm_mean

    """
    # 相当于alpha=2/(1+timeperiod)
    return close.ewm_mean(span=timeperiod, adjust=False, min_periods=timeperiod)


def KAMA(close: Expr, timeperiod: int = 30) -> Expr:
    raise


def MIDPOINT(close: Expr, timeperiod: int = 14) -> Expr:
    return (MAX(close, timeperiod) + MIN(close, timeperiod)) / 2


def MIDPRICE(high: Expr, low: Expr, timeperiod: int = 14) -> Expr:
    return (MAX(high, timeperiod) + MIN(low, timeperiod)) / 2


def RMA(close: Expr, timeperiod: int = 30) -> Expr:
    """TA-Lib does not provide this algorithm explicitly, it is just put here for convenience
    TA-Lib没有明确的提供此算法，这里只是为了调用方便而放在此处

    References
    ----------
    https://pola-rs.github.io/polars/py-polars/html/reference/expressions/api/polars.Expr.ewm_mean.html#polars.Expr.ewm_mean
    https://github.com/twopirllc/pandas-ta/blob/main/pandas_ta/overlap/rma.py

    """
    return close.ewm_mean(alpha=1 / timeperiod, adjust=False, min_periods=timeperiod)


def TEMA(close: Expr, timeperiod: int = 30) -> Expr:
    """

    Notes
    -----
    todo: if the nesting level is too deep, maybe call talib.TEMA directly
    嵌套层数过多，也许直接调用talib.TEMA更快

    """
    EMA1 = EMA(close, timeperiod)
    EMA2 = EMA(EMA1, timeperiod)
    EMA3 = EMA(EMA2, timeperiod)
    return (EMA1 - EMA2) * 3 + EMA3


def TRIMA(close: Expr, timeperiod: int = 30) -> Expr:
    SMA1 = SMA(close, ceil(timeperiod / 2))
    return SMA(SMA1, floor(timeperiod / 2) + 1)



import numpy as np
from polars import Expr, Series, mean_horizontal
from polars import reduce, max_horizontal, sum_horizontal, min_horizontal, Int64


def abs_(x: Expr) -> Expr:
    if isinstance(x, (Expr, Series)):
        return x.abs()
    else:
        return np.abs(x)


def add(a: Expr, b: Expr, *args, filter_: bool = False) -> Expr:
    """Add all inputs (at least 2 inputs required). If filter = true, filter all input NaN to 0 before adding"""
    if filter_:
        return sum_horizontal(a, b, *args)
    _args = [a, b] + list(args)
    return reduce(function=lambda acc, x: acc + x, exprs=_args)


def arc_cos(x: Expr) -> Expr:
    """If -1 <= x <= 1: arccos(x); else NaN"""
    return x.arccos()


def arc_sin(x: Expr) -> Expr:
    """If -1 <= x <= 1: arcsin(x); else NaN"""
    return x.arcsin()


def arc_tan(x: Expr) -> Expr:
    """This operator does inverse tangent of input. """
    return x.arctan()


def ceiling(x: Expr) -> Expr:
    return x.ceil()


def cos(x: Expr) -> Expr:
    return x.cos()


def cosh(x: Expr) -> Expr:
    return x.cosh()


def densify(x: Expr) -> Expr:
    raise


def divide(x: Expr, y: Expr) -> Expr:
    return x / y


def exp(x: Expr) -> Expr:
    return x.exp()


def expm1(x: Expr) -> Expr:
    """convert log return to simple return
    对数收益率 转 简单收益率"""
    return x.exp() - 1


def floor(x: Expr) -> Expr:
    return x.floor()


def fraction(x: Expr) -> Expr:
    """This operator removes the whole number part and returns the remaining fraction part with sign."""
    # 按小学时的定义，负数-1.2的整数部分是-2,小数部分是0.8，而这有所不同
    # return x % 1
    return x.sign() * (x.abs() % 1)


def inverse(x: Expr) -> Expr:
    """1 / x"""
    return 1 / x


def log(x: Expr) -> Expr:
    if isinstance(x, (Expr, Series)):
        return x.log()
    else:
        return np.log(x)


def log10(x: Expr) -> Expr:
    return x.log10()


def log1p(x: Expr) -> Expr:
    """convert simple return to log return
    简单收益率 转 对数收益率"""
    return x.log1p()


def max_(a: Expr, b: Expr, *args) -> Expr:
    """Maximum value of all inputs. At least 2 inputs are required."""
    return max_horizontal(a, b, *args)


def mean(a: Expr, b: Expr, *args) -> Expr:
    return mean_horizontal(a, b, *args)


def min_(a: Expr, b: Expr, *args) -> Expr:
    """Maximum value of all inputs. At least 2 inputs are required."""
    return min_horizontal(a, b, *args)


def mod(x: Expr, y: Expr) -> Expr:
    return x % y


def multiply(a: Expr, b: Expr, *args, filter_: bool = False) -> Expr:
    """Multiply all inputs. At least 2 inputs are required. Filter sets the NaN values to 1"""
    _args = [a, b] + list(args)
    if filter_:
        _args = [_.fill_null(1) for _ in args]

    return reduce(function=lambda acc, x: acc * x, exprs=_args)


def power(x: Expr, y: Expr) -> Expr:
    """x ^ y"""
    return x.pow(y)


def reverse(x: Expr) -> Expr:
    """- x"""
    return -x


def round_(x: Expr, decimals: int = 0) -> Expr:
    """Round input to closest integer."""
    return x.round(decimals)


def round_down(x: Expr, f: int = 1) -> Expr:
    """Round input to greatest multiple of f less than input;"""
    if f == 1:
        return x // 1
    else:
        return x // f * f


def s_log_1p(x: Expr) -> Expr:
    return x.abs().log1p() * x.sign()


def sigmoid(a: Expr) -> Expr:
    # a<0
    # b = a.exp()
    # return b / (1 + b)

    # a>0
    return 1 / (1 + (-a).exp())


def sign(x: Expr) -> Expr:
    if isinstance(x, (Expr, Series)):
        return x.sign()
    else:
        return np.sign(x)


def signed_power(x: Expr, y: Expr) -> Expr:
    """x raised to the power of y such that final result preserves sign of x."""
    if isinstance(y, (int, float)):
        if y == 1:
            return x.abs() * x.sign()
        elif y == 0:
            return x.sign()

    return x.abs().pow(y) * x.sign()


def sin(x: Expr) -> Expr:
    return x.sin()


def sinh(x: Expr) -> Expr:
    return x.sinh()


def softsign(x: Expr) -> Expr:
    return x / (1 + x.abs())


def sqrt(x: Expr) -> Expr:
    return x.sqrt()


def subtract(a: Expr, b: Expr, *args, filter_: bool = False) -> Expr:
    """x-y. If filter = true, filter all input NaN to 0 before subtracting"""
    _args = [a, b] + list(args)
    if filter_:
        _args = [_.fill_null(0) for _ in _args]

    return reduce(function=lambda acc, x: acc - x, exprs=_args)


def tan(x: Expr) -> Expr:
    return x.tan()


def tanh(x: Expr) -> Expr:
    """Hyperbolic tangent of x"""
    return x.tanh()


def truncate(x: Expr) -> Expr:
    """truncate towards zero
    向零取整"""
    return x.cast(Int64)


def var(a: Expr, b: Expr, *args) -> Expr:
    """多列水平方差"""
    _args = [a, b] + list(args)
    _mean = mean_horizontal(_args)
    _sum = sum_horizontal([(expr - _mean) ** 2 for expr in _args])
    return _sum


def std(a: Expr, b: Expr, *args) -> Expr:
    """多列水平标准差"""
    return var(a, b, *args).sqrt()



from polars_ta.wq.arithmetic import *  # noqa
from polars_ta.wq.cross_sectional import *  # noqa
from polars_ta.wq.logical import *  # noqa
from polars_ta.wq.preprocess import *  # noqa
from polars_ta.wq.time_series import *  # noqa
from polars_ta.wq.transformational import *  # noqa
from polars_ta.wq.vector import *  # noqa



from polars import Expr

from polars_ta import TA_EPSILON
from polars_ta.ta.price import MEDPRICE
from polars_ta.tdx.reference import COUNT
from polars_ta.tdx.reference import MA
from polars_ta.tdx.reference import MAX
from polars_ta.tdx.reference import REF
from polars_ta.tdx.reference import SUM


def BRAR_AR(OPEN: Expr, HIGH: Expr, LOW: Expr, CLOSE: Expr, N: int = 26) -> Expr:
    """
    BR:SUM(MAX(0,HIGH-REF(CLOSE,1)),N)/SUM(MAX(0,REF(CLOSE,1)-LOW),N)*100;
    AR:SUM(HIGH-OPEN,N)/SUM(OPEN-LOW,N)*100;

    """

    AR = SUM(HIGH - OPEN, N) / (SUM(OPEN - LOW, N) + TA_EPSILON)  # * 100
    return AR


def BRAR_BR(OPEN: Expr, HIGH: Expr, LOW: Expr, CLOSE: Expr, N: int = 26) -> Expr:
    """
    BR:SUM(MAX(0,HIGH-REF(CLOSE,1)),N)/SUM(MAX(0,REF(CLOSE,1)-LOW),N)*100;
    AR:SUM(HIGH-OPEN,N)/SUM(OPEN-LOW,N)*100;

    """
    LC = REF(CLOSE, 1)
    BR = SUM(MAX(0, HIGH - LC), N) / (SUM(MAX(0, LC - LOW), N) + TA_EPSILON)  # * 100
    return BR


def CR(HIGH: Expr, LOW: Expr, N: int = 26) -> Expr:
    """
    MID:=REF(HIGH+LOW,1)/2;
    CR:SUM(MAX(0,HIGH-MID),N)/SUM(MAX(0,MID-LOW),N)*100;
    MA1:REF(MA(CR,M1),M1/2.5+1);
    MA2:REF(MA(CR,M2),M2/2.5+1);
    MA3:REF(MA(CR,M3),M3/2.5+1);
    MA4:REF(MA(CR,M4),M4/2.5+1);

    """
    MID = REF(MEDPRICE(HIGH, LOW), 1)
    return SUM(MAX(0, HIGH - MID), N) / (SUM(MAX(0, MID - LOW), N) + TA_EPSILON)  # *100


def PSY(CLOSE: Expr, N: int = 12) -> Expr:
    """
    PSY:COUNT(CLOSE>REF(CLOSE,1),N)/N*100;
    PSYMA:MA(PSY,M);

    """
    return COUNT(CLOSE > REF(CLOSE, 1), N) / N


def MASS(HIGH: Expr, LOW: Expr, N1: int = 9, N2: int = 25) -> Expr:
    """
    MASS:SUM(MA(HIGH-LOW,N1)/MA(MA(HIGH-LOW,N1),N1),N2);
    MAMASS:MA(MASS,M);

    """
    MHL = MA(HIGH - LOW, N1)
    return SUM(MHL / MA(MHL, N1), N2)
